# 2014. 12. 9 by Hans Roh hansroh@gmail.com

__version__ = "0.26.18b40"
version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  __version__.split (".")))
NAME = "Skitai/%s.%s" % version_info [:2]

import threading
import sys, os
import h2
from aquests.lib.attrdict import AttrDict
from aquests.protocols.dns import asyndns
from importlib import machinery
from aquests.dbapi import DB_PGSQL, DB_POSTGRESQL, DB_SQLITE3, DB_REDIS, DB_MONGODB

PROTO_HTTP = "http"
PROTO_HTTPS = "https"
PROTO_WS = "ws"
PROTO_WSS = "wss"

WEBSOCKET_SIMPLE = 1
WEBSOCKET_GROUPCHAT = 5

WS_SIMPLE = 1
WS_THREADSAFE = 6
WS_GROUPCHAT = 5

WS_EVT_INIT = "init"
WS_EVT_OPEN = "open"
WS_EVT_CLOSE = "close"
WS_EVT_NONE = None

WS_MSG_JSON = "json"
WS_MSG_XMLRPC = "xmlrpc"
WS_MSG_GRPC = "grpc"
WS_MSG_TEXT = "text"
WS_MSG_DEFAULT = "text"

WS_OPCODE_TEXT = 0x1
WS_OPCODE_BINARY = 0x2
WS_OPCODE_CLOSE = 0x8
WS_OPCODE_PING = 0x9
WS_OPCODE_PONG = 0xa


class _WASPool:
	def __init__ (self):
		self.__wasc = None
		self.__p = {}
		
	def __get_id (self):
		return id (threading.currentThread ())
	
	def __repr__ (self):
		return "<class skitai.WASPool at %x>" % id (self)
			
	def __getattr__ (self, attr):
		_was = self._get ()
		if not _was.in__dict__ ("app") and hasattr (_was, 'request'):
			# it will be called WSGI middlewares except Saddle,
			# So request object not need
			del _was.request			
		return  getattr (_was, attr)
			
	def __setattr__ (self, attr, value):
		if attr.startswith ("_WASPool__"):
			self.__dict__[attr] = value
		else:	
			setattr (self.__wasc, attr, value)
			for _id in self.__p:
				setattr (self.__p [_id], attr, value)
	
	def __delattr__ (self, attr):
		delattr (self.__wasc, attr)
		for _id in self.__p:
			delattr (self.__p [_id], attr, value)
	
	def _start (self, wasc):
		self.__wasc = wasc
	
	def _del (self):
		_id = self.__get_id ()
		try:
			del self.__p [_id]
		except KeyError:
			pass

	def _get (self):
		_id = self.__get_id ()
		try:
			return self.__p [_id]
		except KeyError:
			_was = self.__wasc ()
			self.__p [_id] = _was
			return _was


was = _WASPool ()
def start_was (wasc):
	global was
	was._start (wasc)
		
#------------------------------------------------
# Configure
#------------------------------------------------
dconf = {'mount': {"default": []}, 'clusters': {}, 'cron': [], 'max_ages': {}, 'log_off': [], 'dns_protocol': 'tcp'}

def pref (preset = False):
	from .saddle.Saddle import Config
	
	d = AttrDict ()
	d.config = Config (preset)
	return d

def get_proc_title ():
	a, b = os.path.split (os.path.join (os.getcwd (), sys.argv [0]))
	return "skitai(%s/%s)" % (
		os.path.basename (a),
		b.split(".")[0]
	)
	
def getswd ():
	return os.path.dirname (os.path.join (os.getcwd (), sys.argv [0]))

def is_devel ():
	return not os.environ.get ('SKITAI_PRODUCTION')

def joinpath (*pathes):
	return os.path.normpath (os.path.join (getswd (), *pathes))
abspath = joinpath

Win32Service = None
def set_service (service_class):
	global Win32Service	
	Win32Service = service_class

def set_worker_critical_point (cpu_percent = 90.0, continuous = 3, interval = 20):
	from .server.http_server import http_server
	from .server.https_server import https_server	
	
	http_server.critical_point_cpu_overload = https_server.critical_point_cpu_overload = cpu_percent
	http_server.critical_point_continuous = https_server.critical_point_continuous = continuous
	http_server.maintern_interval = https_server.maintern_interval = interval

def log_off (*path):		
	global dconf
	for each in path:
		dconf ['log_off'].append (each)	

def set_dns_protocol (protocol = 'tcp'):		
	global dconf
	dconf ['dns_protocol'] = protocol
	
def set_max_age (path, max_age):
	global dconf
	dconf ["max_ages"][path] = max_age

def set_max_rcache (objmax):
	global dconf
	dconf ["rcache_objmax"] = objmax

def set_keep_alive (timeout):	
	global dconf
	dconf ["keep_alive"] = timeout

def set_backend_keep_alive (timeout):	
	global dconf
	dconf ["backend_keep_alive"] = timeout

def set_proxy_keep_alive (channel = 60, tunnel = 600):	
	from .server.handlers import proxy

	proxy.PROXY_KEEP_ALIVE = channel
	proxy.PROXY_TUNNEL_KEEP_ALIVE = tunnel
		
def set_request_timeout (timeout):
	global dconf	
	dconf ["network_timeout"] = timeout
set_network_timeout = set_request_timeout

def deflu (*key):
	if "models-keys" not in dconf:
		dconf ["models-keys"] = []
	if isinstance (key [0], (list, tuple)):
		key = list (key [0])
	dconf ["models-keys"].extend (key)
addlu = trackers = lukeys = deflu

def mount (point, target, appname = "app", pref = pref (True), host = "default", path = None):
	global dconf
	
	def init_app (modpath, pref):
		modinit = os.path.join (os.path.dirname (modpath), "__init__.py")
		if os.path.isfile (modinit):
			loader = machinery.SourceFileLoader('temp', modinit)
			mod = loader.load_module()
			hasattr (mod, "bootstrap") and mod.bootstrap (pref)
	
	if path:
		if isinstance (path, str):
			path = [path]
		path.reverse ()	
		for each in path:			
			sys.path.insert (0, abspath (each))
			
	if hasattr (target, "__file__"):
		target = (target, '__export__.py')
	
	if type (target) is tuple: 
		module, appfile = target
		target = os.path.join (os.path.dirname (module.__file__), "export", "skitai", appfile)
			
	if type (target) is not str:
		# app instance
		target = os.path.join (os.getcwd (), sys.argv [0])
	else:
		if target [0] != "@":		
			target = os.path.join (getswd (), target)
	
	if host not in dconf ['mount']:
		dconf ['mount'][host] = []
	if os.path.isdir (target):
		dconf ['mount'][host].append ((point, target, None))		
	else:	
		init_app (target, pref)
		if appname:
			app = (target, appname)
		else:
			app = target
		dconf ['mount'][host].append ((point, app, pref))
	
def cron (sched, cmd):
	global dconf
	dconf ["cron"].append ('%s %s' % (sched, cmd))

def alias (name, ctype, members, role = "", source = "", ssl = False):
	from .server.rpc.cluster_manager import AccessPolicy
	global dconf	
	if name [0] == "@":
		name = name [1:]
	if dconf ["clusters"].get (name):
		return
	policy = AccessPolicy (role, source)
	args = (ctype, members, policy, ssl)
	dconf ["clusters"][name] = args

def enable_forward (forward_to = 443, port = 80, ip = ""):
	global dconf
	dconf ['fws_address'] = ip
	dconf ['fws_port'] = forward_to
	dconf ['fws_to'] = port

def enable_gateway (enable_auth = False, secure_key = None, realm = "Skitai API Gateway"):
	global dconf
	dconf ["enable_gw"] = True
	dconf ["gw_auth"] = enable_auth,
	dconf ["gw_realm"] = realm,
	dconf ["gw_secret_key"] = secure_key

def _get_django_settings (settings_path):
	import importlib
	import django
	
	ap = abspath (settings_path)
	django_main, settings_file = os.path.split (ap)
	django_root, django_main_dir = os.path.split (django_main)
	settings_mod = "{}.{}".format (django_main_dir, settings_file.split (".")[0])
	
	if not os.environ.get ("DJANGO_SETTINGS_MODULE"):		
		sys.path.insert (0, django_root)	
		os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings_mod)
		django.setup()
	
	return importlib.import_module(settings_mod).DATABASES
	
def use_django_models (settings_path, _alias = None):
	dbsettings = _get_django_settings (settings_path)
	if _alias:
		default = dbsettings ['default']
		if default ['ENGINE'].endswith ('sqlite3'):			
			alias (_alias, DB_SQLITE3, default ['NAME'])
		else:
			if not default.get ("PORT"):
				default ["PORT"] = 5432
			if not default.get ("HOST"):
				default ["HOST"] = "127.0.0.1"
			if not default.get ("USER"):
				default ["USER"] = ""
			if not default.get ("PASSWORD"):
				default ["PASSWORD"] = ""
			alias (_alias, DB_PGSQL, "%(HOST)s:%(PORT)s/%(NAME)s/%(USER)s/%(PASSWORD)s" % default)
			
	return dbsettings
	
def mount_django (point, wsgi_path, pref = pref (True), dbalias = None, host = "default"):
	path = os.path.dirname (os.path.dirname (wsgi_path))
	mount (point, wsgi_path, "application", pref, host, path)
	if dbalias:
		return use_django_models (
			os.path.join (os.path.dirname (wsgi_path), 'settings.py'),
			dbalias
		)
	
def enable_cachefs (memmax = 0, diskmax = 0, path = None):
	global dconf	
	dconf ["cachefs_memmax"] = memmax
	dconf ["cachefs_dir"] = path
	dconf ["cachefs_diskmax"] = diskmax	
				
def enable_proxy (unsecure_https = False):
	global dconf
	dconf ["proxy"] = True
	dconf ["proxy_unsecure_https"] = unsecure_https
	if os.name == "posix": 
		dconf ['dns_protocol'] = 'udp'

def enable_blacklist (path):
	global dconf
	dconf ["blacklist_dir"] = path

def enable_ssl (certfile, keyfile, passphrase):
	global dconf			
	dconf ["certfile"] = certfile
	dconf ["keyfile"] = keyfile
	dconf ["passphrase"] = passphrase
	
def enable_smtpda (server = None, user = None, password = None, ssl = None, max_retry = None, keep_days = None):
	global dconf
	smtpda = {}
	if server: smtpda ["server"] = server
	if user: smtpda ["user"] = user
	if password: smtpda ["password"] = password
	if ssl: smtpda ["ssl"] = ssl
	if max_retry: smtpda ["max-retry"] = max_retry
	if keep_days: smtpda ["keep-days"] = keep_days
	dconf ["smtpda"] = smtpda

def run (**conf):
	import os, sys, time
	from . import lifetime
	from .server import Skitai	
	from .server.wastuff import process, daemon	as wasdaemon
	from aquests.lib.pmaster import flock
	import getopt
		
	class SkitaiServer (Skitai.Loader):
		NAME = 'instance'
		
		def __init__ (self, conf):
			self.conf = conf
			self.children = []
			self.flock = None
			Skitai.Loader.__init__ (self, 'config', conf.get ('logpath'), self.get_varpath ())
			
		def close (self):
			if self.wasc.httpserver.worker_ident == "master":
				if self.children:
					for child in self.children:					
						self.wasc.logger ("server", "[info] killing %s..." % child.name)						
						child.kill (0)
					
					for i in range (10):
						time.sleep (1)
						veto = False
						for child in self.children:
							veto = (child.poll () is None)
							if veto:
								if i % 3 == 0:
									self.wasc.logger ("server", "[info] %s is still alive" % child.name)
								break
						if not veto:
							break
					
					if veto:
						for child in self.children:
							if child.poll () is None:
								self.wasc.logger ("server", "[info] force to kill %s..." % child.name)
								child.kill (1)
			
			Skitai.Loader.close (self)
			
		def create_process (self, name, args, karg):
			argsall = []
			karg ['pname'] = get_proc_title ()
			karg ['var-path'] = self.get_varpath ()
			if self.conf.get ("logpath"):
				karg ['log-path'] = self.conf.get ("logpath")		
			if self.conf.get ("verbose", "no") in ("yes", 1, "1"):
				karg ['verbose'] = "yes"
			
			if karg:
				for k, v in karg.items ():
					if len (k) == 1:
						h = "-"
					else:
						h = "--"
					if v is None:
						argsall.append ('%s%s' % (h, k))
					else:
						argsall.append ('%s%s "%s"' % (h, k, str (v).replace ('"', '\\"')))
			
			if args:
				for each in args:
					argsall.append ('"%s"' % each.replace ('"', '\\"'))
								
			cmd = "%s %s %s" % (
				sys.executable, 
				os.path.join (os.path.dirname (os.path.dirname (Skitai.__file__)), "bin", name + ".py"), 
				" ".join (argsall)
			)
			
			self.children.append (
				process.Process (
					cmd, 
					name,
					self.get_varpath ()
				)
			)
		
		def get_varpath (self):
			return self.conf.get ('varpath') or wasdaemon.get_default_varpath ()
			
		def config_logger (self, path):
			media = []
			if path is not None:
				media.append ("file")
			if self.conf.get ('verbose', "no") in ("yes", "1", 1):				
				media.append ("screen")
				self.conf ['verbose'] = "yes"
			if not media:
				media.append ("screen")
				self.conf ['verbose'] = "yes"
			Skitai.Loader.config_logger (self, path, media, self.conf ["log_off"])
		
		def master_jobs (self):
			smtpda = self.conf.get ('smtpda')
			if smtpda is not None:				
				self.create_process ('smtpda', [], smtpda)			
			cron = self.conf.get ('cron')
			if cron:
				self.create_process ('cron', cron, {})
			
			self.wasc.logger ("server", "[info] active var dir %s" % self.get_varpath ())
			if self.logpath:
				self.wasc.logger ("server", "[info] logging to %s" % self.logpath)
			
			self.conf.get ("models-keys") and self.set_model_keys (self.conf ["models-keys"])
						
		def maintern_shutdown_request (self, now):
			req = self.flock.lockread ("signal")
			if not req: return
			self.wasc.logger ("server", "[info] got signal - %s" % req)
			if req == "terminate":			
				lifetime.shutdown (0, 30.0)
			elif req == "restart":			
				lifetime.shutdown (3, 30.0)
			elif req == "kill":
				lifetime.shutdown (0, 1.0)
			elif req == "rotate":
				self.wasc.logger.rotate ()
			else:
				self.wasc.logger ("server", "[error] unknown signal - %s" % req)
			self.flock.unlock ("signal")
				
		def configure (self):
			conf = self.conf
			self.set_num_worker (conf.get ('workers', 1))
			if conf.get ("certfile"):
				self.config_certification (conf.get ("certfile"), conf.get ("keyfile"), conf.get ("passphrase"))
			
			self.config_dns (dconf ['dns_protocol'])
			if conf.get ("cachefs_diskmax", 0) and not conf.get ("cachefs_dir"):
				conf ["cachefs_dir"] = os.path.join (self.get_varpath (), "cachefs")

			self.config_cachefs (
				conf.get ("cachefs_dir", None), 
				conf.get ("cachefs_memmax", 0),
				conf.get ("cachefs_diskmax", 0)				
			)
			self.config_rcache (conf.get ("rcache_objmax", 100))
			if conf.get ('fws_to'):
				self.config_forward_server (
					conf.get ('fws_address', '0.0.0.0'), conf.get ('fws_port', 80), conf.get ('fws_to', 443)
				)
			
			# before fork
			self.master_jobs ()
				
			self.config_webserver (
				conf.get ('port', 5000), conf.get ('address', '0.0.0.0'),
				NAME, conf.get ("certfile") is not None,
				conf.get ('keep_alive', 30), 
				conf.get ('network_timeout', 30)
			)
			
			if os.name == "posix" and self.wasc.httpserver.worker_ident == "master":
				# master does not serve
				return			
			
			self.config_threads (conf.get ('threads', 4))			
			self.config_backends (conf.get ('backend_keep_alive', 1200))
			for name, args in conf.get ("clusters", {}).items ():				
				ctype, members, policy, ssl = args
				self.add_cluster (ctype, name, members, ssl, policy)
			
			self.install_handler (
				conf.get ("mount"), 
				conf.get ("proxy", False),
				conf.get ("max_ages", {}),
				conf.get ("blacklist_dir"), # blacklist_dir
				conf.get ("proxy_unsecure_https", False), # disable unsecure https
				conf.get ("enable_gw", False), # API gateway
				conf.get ("gw_auth", False),
				conf.get ("gw_realm", "API Gateway"),
				conf.get ("gw_secret_key", None)
			)
			
			lifetime.init (logger = self.wasc.logger.get ("server"))
			lifetime.maintern.sched (3.0, asyndns.pool.maintern)
			if os.name == "nt":
				lifetime.maintern.sched (11.0, self.maintern_shutdown_request)								
				self.flock = flock.Lock (os.path.join (self.get_varpath (), ".%s" % self.NAME))
			
	#----------------------------------------------------------------------------
	if os.name == "nt":
		import win32serviceutil
		
		def set_service_config (argv = []):
			global Win32Service			
			#sys.stdout = sys.stderr = open (r"D:\apps\skitai\examples\err.log", "a")
			argv.insert (0, "")			
			script = os.path.join (os.getcwd (), sys.argv [0])
			win32serviceutil.HandleCommandLine(Win32Service, "%s.%s" % (script [:-3], Win32Service.__name__), argv)
				
		def install (working_dir):
			set_service_config (['--startup', 'auto', 'install'])
	
		def remove (working_dir):
			set_service_config (['remove'])
		
		def update (working_dir):
			set_service_config (['update'])
				
	def start (working_dir):
		if os.name == "nt":
			set_service_config (['start'])
		else:	
			from aquests.lib.pmaster import Daemonizer
			if not Daemonizer (working_dir, 'skitai').runAsDaemon ():
				print ("already running")
				sys.exit ()
			
	def stop (working_dir):
		if os.name == "nt":			
			set_service_config (['stop'])
		else:	
			from aquests.lib.pmaster import daemon
			daemon.kill (working_dir, 'skitai', True)
	
	def status (working_dir):
		from aquests.lib.pmaster import daemon
		pid = daemon.status (working_dir, 'skitai')
		if pid:
			print ("running [%d]" % pid)
		else:
			print ("stopped")
	
	#----------------------------------------------------------------------------
	
	global dconf
	
	for k, v in dconf.items ():
		if k not in conf:
			conf [k] = v
				
	if not conf.get ('mount'):
		raise systemError ('No mount point')
	
	argopt = getopt.getopt(sys.argv[1:], "vfdsr", [])
	working_dir = getswd ()
	try: cmd = argopt [1][0]
	except: cmd = None
	karg = {}
	for k, v in argopt [0]:
		if k == "-d": cmd = "start"
		elif k == "-r": cmd = "retart"
		elif k == "-s": cmd = "stop"
		else: karg [k] = v
			
	if cmd in ("start", "restart") and '-v' in karg:
		raise SystemError ('Daemonizer cannot be run with -v, It is meaningless')
				
	if cmd == "stop":
		return stop (working_dir)
	elif cmd == "status":
		return status (working_dir)
	elif cmd == "start":
		start (working_dir)		
	elif cmd == "install":	
		install (working_dir)		
	elif cmd == "update":	
		update (working_dir)
	elif cmd == "remove":	
		remove (working_dir)	
	elif cmd == "restart":
		stop (working_dir)		
		time.sleep (2)
		start (working_dir)	
	elif cmd:
		print ('unknown command: %s' % cmd)
		return
				
	if cmd and os.name == "nt":
		return
				
	verbose = 0
	if '-v' in karg or conf.get ('logpath') is None:
		verbose = 1
	
	if "-f" in karg:
		# failsafe run
		from . import failsafer
		failsafer.Service (
			"%s %s %s" % (sys.executable, os.path.join (os.getcwd (), sys.argv [0]), verbose and '-v' or ''),
			conf.get ('logpath'),
			conf.get ('varpath'),
			'-v' in karg
		).run ()
	
	else:
		os.chdir (working_dir)
		if verbose:
			conf ['verbose'] = 'yes'
		server = SkitaiServer (conf)
		# timeout for fast keyboard interrupt on win32	
		try:
			try:
				server.run (conf.get ('verbose') and 2.0 or 30.0)
			except KeyboardInterrupt:
				pass	
		
		finally:	
			_exit_code = server.get_exit_code ()
			if _exit_code is not None: # master process				
				sys.exit (_exit_code)
			else: 
				# worker process				
				# for avoiding multiprocessing.manager process's join error
				# use os._exit
				os._exit (lifetime._exit_code)
			



import zlib
import time
import os
import sys
from aquests.protocols.http import http_date, http_util
from aquests.lib.reraise import reraise 
from aquests.lib import producers, compressors
from aquests.protocols.http import respcodes
from .wastuff import selective_logger
from .wastuff.api import API
from .wastuff.api import catch
import skitai
import asyncore
import json
from ..saddle.exceptions import HTTPError

try: 
	from urllib.parse import urljoin
except ImportError:
	from urlparse import urljoin	

UNCOMPRESS_MAX = 2048
ONETIME_COMPRESS_MAX = 1048576

# Default error message
DEFAULT_ERROR_MESSAGE = """<!DOCTYPE html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<title>%(code)d %(message)s</title>
<style type="text/css"><!-- *{font-family:sans-serif, verdana;}body{margin:0;padding:0;font-size:14px;color:#1e1e1e;font-family:verdana,sans-serif;} #titles{margin-left:16px;}#titles h1{color: #000000;} #content{padding:16px 16px 30px 16px;} #debug h2 {font-size: 18px; font-weight: 600;} #debug h3{font-size: 16px; color: #d90000;} #debug p,b,h4,li {font-size:14px;}#debug h4{color: #999999;} #debug li{margin-bottom: 6px;} #debug .f {color:#8AB088; font-weight: bold;} #debug .n {color:#003366;font-weight:bold;} #debug{background-color: #E9F5F9; ; margin-bottom: 32px; padding:8px 16px 8px 16px; border-radius: 6px;} hr{margin:0;padding:0} #debug #debug li,i{font-weight:normal;}#footer {font-size:12px;padding-left:10px;}#software div {padding-top: 10px; font-size: 16px;} --></style>
</head>
<body>
<div id="titles"><h1>%(code)d %(message)s</h1></div>
<div id="content">
<div id="%(mode)s"><p>%(traceback)s</p></div>
<div id="software"><hr noshade size="1"><div>%(software)s</div></div>
</div>
</body>
</html>"""


class http_response:
	log_or_not = selective_logger.SelectiveLogger ()
	reply_code = 200
	reply_message = "OK"
	_is_async_streaming = False
		
	def __init__ (self, request):
		self.request = request
		self.reply_headers = [
			('Server', skitai.NAME),
			('Date', http_date.build_http_date (time.time()))
		]
		self.outgoing = producers.fifo ()
		self._is_done = False
		self.stime = time.time ()
		self.htime = 0
		self.content_type = None
		self.current_app = None
		self.current_env = None
		
	def is_async_streaming (self):
		return self._is_async_streaming
	
	def set_streaming (self):
		self._is_async_streaming = True	
	
	def is_responsable (self):
		return not self._is_done
	
	def is_done (self):	
		return self._is_done
				
	def __len__ (self):
		return len (self.outgoing)
			
	def __setitem__ (self, key, value):
		self.set (key, value)

	def __getitem__ (self, key):
		return self.get (key)
	
	def __delitem__ (self, k):
		self.delete (k)
	
	def has_key (self, key):
		key = key.lower ()
		return key in [x [0].lower () for x in self.reply_headers]		
			
	def set (self, key, value):
		self.reply_headers.append ((key, value))
		
	def get (self, key):
		key = key.lower ()
		for k, v in self.reply_headers:
			if k.lower () == key:
				return v
			
	def delete (self, key):
		index = 0
		found = 0
		key = key.lower ()
		for hk, hv in self.reply_headers:
			if key == hk.lower ():
				found = 1
				break
			index += 1
		
		if found:
			del self.reply_headers [index]
			self.delete (key)
		
	def update (self, key, value):
		self.delete (key)
		self.set (key, value)
	
	def append_header (self, key, value):
		val = self.get (key)
		if not val:
			self.set (key, value)
		else:
			self.set (key, val + ", " + value)

	def build_reply_header (self, with_header = 1):
		h = [self.response (self.reply_code, self.reply_message)]
		if with_header:
			h.extend (['%s: %s' % x for x in self.reply_headers])
		h = '\r\n'.join (h) + '\r\n\r\n'			
		return h		
	
	def get_status_msg (self, code):
		return respcodes.get (code, "Undefined Error")
		
	def response (self, code, status):	
		return 'HTTP/%s %d %s' % (self.request.version, code, status)
	
	def parse_ststus (self, status):
		try:	
			code, status = status.split (" ", 1)
			code = int (code)
		except:
			raise AssertionError ("Can't understand given status code")		
		return code, status	

	#--------------------------------------------		
	# for WSGI & Saddle Apps
	#--------------------------------------------		
	def set_reply (self, status):
		# for Saddle App
		self.reply_code, self.reply_message = self.parse_ststus (status)
	
	def get_reply (self):
		# for Saddle App
		return self.reply_code, self.reply_message
	
	def instant (self, status = "", headers = None):
		# instance messaging		
		if self.request.channel is None:
			return
		code, msg = self.parse_ststus (status)
		reply = [self.response (code, msg)]
		if headers:
			for header in headers:
				reply.append ("%s: %s" % header)
		self.request.channel.push (("\r\n".join (reply) + "\r\n\r\n").encode ("utf8"))
		
	def start_response (self, status, headers = None, exc_info = None):		
		# for WSGI App
		if not self.is_responsable ():
			if exc_info:
				try:
					reraise (*exc_info)
				finally:
					exc_info = None	
			else:
				raise AssertionError ("Response already sent!")		
			return
		
		code, status = self.parse_ststus (status)
		self.start (code, status, headers)
		
		if exc_info:
			# expect plain/text, send exception info to developers
			content = catch (0, exc_info)
			self.push (content)
			
		return self.push #by WSGI Spec.
	
	#----------------------------------------	
	# Internal Rsponse Methods.
	#----------------------------------------	
	def abort (self, code, status = "", why = ""):
		self.request.channel.reject ()
		self.error (code, status, why, force_close = True)
		
	def start (self, code, status = "", headers = None):
		if not self.is_responsable (): return
		self.reply_code = code
		if status: self.reply_message = status
		else:	self.reply_message = self.get_status_msg (code)			
		if headers:
			for k, v in headers:
				self.set (k, v)
	reply = start
	
	def catch (self, exc_info = None):
		if self.request.get_header ('accept').find ("application/json") != -1:
			return catch (2, exc_info)
		return catch (1, exc_info)
		
	def build_error_template (self, why = ''):
		global DEFAULT_ERROR_MESSAGE
		
		exc_info = None
		if type (why) is tuple: # sys.exc_info ()
			exc_info, why  = why, ''
		
		error = {}
		if self.request.get_header ('accept', '').find ("application/json") != -1:			
			message = why or self.get_status ()
			code = 10000
			debug = None
			if exc_info:
				code = 10001
				debug = 'see traceback' 
			return self.fault (message, code, debug, exc_info = exc_info)
		
		else:	
			self.update ('content-type', 'text/html')
			error ['detail'] = why
			error ['time'] = http_date.build_http_date (time.time ())
			error ['url'] = urljoin ("%s://%s/" % (self.request.get_scheme (), self.request.get_header ("host")), self.request.uri)
			error ['software'] = skitai.NAME			
			error ['mode'] = exc_info and 'debug' or 'normal'				
			error ['code'] = self.reply_code
			error ['message'] = self.reply_message
			error ["traceback"] = exc_info and catch (1, exc_info) or ""
			
			content = None	
			if self.current_app and hasattr (self.current_app, 'get_error_page'):
				try:
					content = self.current_app.get_error_page (error)
				except:							
					self.request.logger.trace ()				
					if self.current_app.debug:						
						error ["traceback"] += "<h2 style='padding-top: 40px;'>Exception Occured During Building Error Template</h2>" + catch (1)
						
			return content or (DEFAULT_ERROR_MESSAGE % error)
				
	def error (self, code, status = "", why = "", force_close = False, push_only = False):
		if not self.is_responsable (): return
		self.outgoing.clear ()
		self.reply_code = code
		if status: self.reply_message = status
		else:	self.reply_message = self.get_status_msg (code)
		
		body = self.build_error_template (why).encode ("utf8")
		self.update ('Content-Length', len(body))				
		self.update ('Cache-Control', 'max-age=0')
		self.push (body)
		if not push_only:
			self.done (force_close)
	
	#--------------------------------------------
	# Send Response
	#--------------------------------------------	
	def die_with (self, thing):
		if self.request.channel:
			self.request.channel.attend_to (thing)
	
	def hint_promise (self, *args, **kargs):
		# ignore in version 1.x
		pass
	
	def push_and_done (self, thing, *args, **kargs):
		self.push (thing)
		self.done (*args, **kargs)
			
	def push (self, thing):		
		if not self.is_responsable (): 
			return
		if type(thing) is bytes:
			self.outgoing.push (producers.simple_producer (thing))
		else:			
			self.outgoing.push (thing)
			    					
	def done (self, force_close = False, upgrade_to = None, with_header = 1):
		self.content_type = self.get ('content-type')
				
		if not self.is_responsable (): return
		self._is_done = True
		if self.request.channel is None: return
			
		self.htime = (time.time () - self.stime) * 1000
		self.stime = time.time () #for delivery time
		
		# compress payload and globbing production
		do_optimize = True
		if upgrade_to or self.is_async_streaming ():
			do_optimize = False
						
		connection = http_util.get_header (http_util.CONNECTION, self.request.header).lower()
		close_it = False
		way_to_compress = ""
		wrap_in_chunking = False
		
		if force_close:
			close_it = True
			if self.request.version == '1.1':
				self.update ('Connection', 'close')
			else:	
				self.delete ('Connection')
				
		else:
			if self.request.version == '1.0':
				if connection == 'keep-alive':
					if not self.has_key ('content-length'):
						close_it = True
						self.update ('Connection', 'close')
					else:
						self.update ('Connection', 'keep-alive')						
				else:
					close_it = True
			
			elif self.request.version == '1.1':
				if connection == 'close':
					close_it = True
					self.update ('Connection', 'close')				
				if not self.has_key ('transfer-encoding') and not self.has_key ('content-length') and self.has_key ('content-type'):
					wrap_in_chunking = True
					
			else:
				# unknown close
				self.update ('Connection', 'close')
				close_it = True
		
		if len (self.outgoing) == 0:
			self.update ('Content-Length', "0")
			self.delete ('transfer-encoding')			
			self.delete ('content-type')
			outgoing_producer = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))
			do_optimize = False
			
		elif len (self.outgoing) == 1 and hasattr (self.outgoing.first (), "ready"):
			outgoing_producer = producers.composite_producer (self.outgoing)
			if wrap_in_chunking:
				self.update ('Transfer-Encoding', 'chunked')
				outgoing_producer = producers.chunked_producer (outgoing_producer)
			outgoing_header = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))
			self.request.channel.push_with_producer (outgoing_header)
			do_optimize = False
			
		elif do_optimize and not self.has_key ('Content-Encoding'):
			maybe_compress = self.request.get_header ("Accept-Encoding")
			if maybe_compress:
				cl = self.has_key ("content-length") and int (self.get ("Content-Length")) or -1
				if cl == -1:
					cl = self.outgoing.get_estimate_content_length ()
					
				if 0 < cl  <= UNCOMPRESS_MAX:
					maybe_compress = ""
				elif not wrap_in_chunking and cl > ONETIME_COMPRESS_MAX:
					# too big for memory, do not compress
					maybe_compress = ""
			
			if maybe_compress:
				content_type = self.get ("Content-Type")
				if content_type and (content_type.startswith ("text/") or content_type.startswith ("application/json")):
					accept_encoding = [x.strip () for x in maybe_compress.split (",")]
					if "gzip" in accept_encoding:
						way_to_compress = "gzip"
					elif "deflate" in accept_encoding:
						way_to_compress = "deflate"
			
			if way_to_compress:
				if self.has_key ('Content-Length'):
					self.delete ("content-length") # rebuild
				self.update ('Content-Encoding', way_to_compress)	
			
			if wrap_in_chunking:
				outgoing_producer = producers.composite_producer (self.outgoing)
				self.delete ('content-length')
				self.update ('Transfer-Encoding', 'chunked')				
				if way_to_compress:
					if way_to_compress == "gzip": 
						compressing_producer = producers.gzipped_producer
					else: # deflate
						compressing_producer = producers.compressed_producer
					outgoing_producer = compressing_producer (outgoing_producer)					
				outgoing_producer = producers.chunked_producer (outgoing_producer)
				outgoing_header = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))				
				
			else:
				self.delete ('transfer-encoding')
				if way_to_compress:					
					if way_to_compress == "gzip":
						compressor = compressors.GZipCompressor ()
					else: # deflate
						compressor = zlib.compressobj (6, zlib.DEFLATED)
					cdata = b""
					has_producer = 1
					while 1:
						has_producer, producer = self.outgoing.pop ()
						if not has_producer: break
						while 1:	
							data = producer.more ()
							if not data:
								break
							cdata += compressor.compress (data)
					cdata += compressor.flush ()					
					self.update ("Content-Length", len (cdata))
					outgoing_producer = producers.simple_producer (cdata)						
				else:
					outgoing_producer = producers.composite_producer (self.outgoing)
					
				outgoing_header = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))				
			
			outgoing_producer = producers.composite_producer (
				producers.fifo([outgoing_header, outgoing_producer])
			)

		outgoing_producer = self.log_or_not (self.request.uri, outgoing_producer, self.log)				
		if do_optimize:
			outgoing_producer = producers.globbing_producer (outgoing_producer)
		
		# IMP: second testing after push_with_producer()->init_send ()
		if self.request.channel is None: return

		if upgrade_to:
			request, terminator = upgrade_to
			self.request.channel.current_request = request
			self.request.channel.set_terminator (terminator)
		else:
			# preapre to receice new request for channel
			self.request.channel.current_request = None
			self.request.channel.set_terminator (b"\r\n\r\n")
		
		# proxy collector and producer is related to asynconnect
		# and relay data with channel
		# then if request is suddenly stopped, make sure close them
		self.die_with (self.request.collector)
		self.die_with (self.request.producer)
		
		logger = self.request.logger #IMP: for  disconnect with request
		try:
			if outgoing_producer:
				self.request.channel.push_with_producer (outgoing_producer)
			if close_it:
				self.request.channel.close_when_done ()
		except:			
			logger.trace ()
						
	def log (self, bytes):
		server = self.request.channel.server		
		referer = self.request.get_header ('referer')
		forwared_for = self.request.get_header ('x-forwarded-for')
		worker_id = server.worker_ident [8:]
		worker = worker_id and ("W" + worker_id) or "M"
			
		server.log_request (
			'%s %s %s %s %s %d %s %d %s %s %s %s %s %s %s %s %d %d %d'
			% (
			self.request.channel.addr[0],
			
			self.request.host or "-",
			self.request.is_promise () and "PUSH" or self.request.method,
			self.request.uri,
			self.request.version,
			self.request.rbytes, # recv body size			
			
			self.reply_code,			
			bytes, # send body size
			
			self.request.gtxid or "-",
			self.request.ltxid or "-",
			self.request.user and '"' + str (self.request.user) + '"' or "-",
			self.request.token or "-",
			
			referer and '"' + referer + '"' or "-",
			self.request.user_agent and '"' + self.request.user_agent + '"' or "-",
			forwared_for or '-',
			
			worker,
			len (asyncore.socket_map),
			self.htime, # due time to request handling
			(time.time () - self.stime) * 1000, # due time to sending data			
			)
		)
		# clearing resources, back refs
		self.request.response_finished ()
	
	#---------------------------------------------
	# Used within Saddle app
	#---------------------------------------------
	
	set_header = set
	get_header = get
	del_header = delete
	
	def set_cache (self, max_age = 0):
		if self.request.version == "1.0":			
			self.set_header ('Expires', http_date.build_http_date (time.time() + max_age))				
		else:
			self.set_header ('Cache-Control', "max-age=%d" % max_age)
					
	def set_headers (self, headers):
		for k, v in headers:
			self.set (k, v)
	
	def append_headers (self, headers):
		for k, v in headers:
			self.set (k, v)		
			
	def get_hedaers (self):	
		return self.reply_headers	
	set_status = set_reply
		
	def get_status (self):
		return "%d %s" % self.get_reply ()
	
	# API methods ------------------------------------------------------------
	def api (self, __data_dict__ = None, **kargs):
		return API (self.request, __data_dict__ or kargs)

	def fault (self, message, code = 20000,  debug = None, more_info = None, exc_info = None):
		api = self.api ()
		api.error (message, code, debug, more_info, exc_info)
		return api
	
	def traceback (self, message = '', code = 10001, debug = 'see traceback', more_info = None):
		api = self.api ()
		api.traceback (message, code, debug, more_info)
		return api
	
	def for_api (self, status = "200 OK", *args, **kargs):
		if status [0] == "2":
			r = self.api (*args, **kargs)
		else:
			r = self.fault (*args, **kargs)
		return self (status, r)
	
	# Returning ------------------------------------------------------------------

	def file (self, path, mimetype = 'application/octet-stream', filename = None):
		self.set_header ('Content-Type',  mimetype)
		self.set_header ('Content-Length', str (os.path.getsize (path)))	
		if filename:
			self.set_header ('Content-Disposition', 'attachment; filename="{}"'.format (filename))
		return producers.file_producer (open (path, "rb"))					
	
	def throw (self, status):
		raise HTTPError (status)
				
	def __call__ (self, status = "200 OK", body = None, headers = None, exc_info = None):
		self.start_response (status, headers)
		if body is None:
			return self.build_error_template (exc_info)
		return body
	
	# will be derecating
	eapi = fault
	
	


import pickle as pickle
from aquests.protocols.http import http_date
import time
import random
from aquests.lib import pathtool
from aquests.lib import udict
import os, sys
try:
	from urllib.parse import quote_plus, unquote_plus
except ImportError:
	from urllib import quote_plus, unquote_plus	
import base64
import pickle
from hmac import new as hmac

_default_hash = None
if sys.version_info >= (2, 5):
	try:
		from hashlib import sha1 as _default_hash
	except ImportError:
		pass
if _default_hash is None:
	import sha as _default_hash


class UnquoteError(Exception):
	"""Internal exception used to signal failures on quoting."""


class SecuredValue:
	hash_method = _default_hash
	serialization_method = pickle
	quote_base64 = True
	KEY = "sdlsecure"	
	
	def __init__ (self, name, cookie, request, secret_key):
		self.name = "_" + name.upper ()
		self.cookie = cookie
		self.request = request
		self.secret_key = secret_key
		self.data = None
		self.dirty = False
		self.__config = None		
		self._source_verified = False		
			
	def __contains__ (self, k):
		self.data is None and self.unserialize ()
		return k in self.data
	
	def _recal_expires (self, expires):						
		return expires
	
	def clear (self):
		self.data is None and self.unserialize ()
		self.set_default_data ()
		self.dirty = True
	
	def validate (self):
		pass
	
	def source_verified (self):
		return self._source_verified
			
	def set_default_data (self):
		self.data = None
	
	def rollback (self):
		self.dirty = False
	
	def config (self, path = None, domain = None, secure = False, http_only = False):
		self.__config = (path, domain, secure, http_only)
	
	def set_cookie (self, expires):
		if expires != 0:
			data = self.serialize ()
		else:
			data = ""	
			
		if self.__config:
			self.cookie.set (self.KEY + self.name, data, expires, *self.__config)
		else:
			self.cookie.set (self.KEY + self.name, data, expires)			
		self.dirty = False
	
	def unserialize (self):	
		string = self.cookie.get_named_session_data (self.KEY + self.name)
		if not string:
			return self.set_default_data ()
		self.unserialize_from_string (string.encode ("utf8"))		
		self.validate ()	
		
	def commit (self, expires = None):
		if not self.dirty or self.data is None: 
			return
		self.set_cookie (self._recal_expires (expires))
		
	def quote (self, value):
		if self.serialization_method is not None:
			value = self.serialization_method.dumps(value, 1)			
		if self.quote_base64:
			value = base64.b64encode (value)
			value = b''.join(value.splitlines()).strip()
		return value

	def unquote(self, value):
		try:
			if self.quote_base64:
				value = base64.b64decode(value)
			if self.serialization_method is not None:
				value = self.serialization_method.loads(value)
			return value
		except:			
			raise UnquoteError


#------------------------------------------------------
# Dict Type
#------------------------------------------------------

class SecuredDictValue (SecuredValue):
		
	def set_default_data (self):
		self.data = {}
		
	def __setitem__ (self, k, v):
		self.set (k, v)
	
	def __delitem__ (self, k):
		return self.remove (k)
	
	def __getitem__ (self, k):
		return self.get (k)
	
	def __iter__ (self):
		self.data is None and self.unserialize ()
		return self.data.__iter ()
		
	def has_key (self, k):
		self.data is None and self.unserialize ()
		return k in self.data
	
	def items (self):
		self.data is None and self.unserialize ()
		return list(self.data.items ())
	
	def keys (self):
		self.data is None and self.unserialize ()
		return list(self.data.keys ())	
	
	def values (self):
		self.data is None and self.unserialize ()
		return list(self.data.values ())	
			
	def remove (self, k):
		self.data is None and self.unserialize ()
		try:
			del self.data [k]
		except KeyError:
			pass
		else:
			self.dirty = True
			
	def set (self, k, v):
		self.data is None and self.unserialize ()
		if type (k) is not type (""):
			raise TypeError("Session key must be string type")
		self.data [k] = v
		self.dirty = True
	
	def get (self, k, v = None):
		self.data is None and self.unserialize ()
		return self.data.get (k, v)
		
	def serialize(self):
		if self.secret_key is None:
			raise RuntimeError('no secret key defined')
		
		result = []
		mac = hmac(self.secret_key, None, self.hash_method)
		for key, value in sorted (self.items(), key = lambda x: x[0]):
			result.append (quote_plus (key).encode ("utf8") + b"=" + self.quote(value))
			mac.update(b'|' + result[-1])			
		return (base64.b64encode(mac.digest()).strip() + b"?" + b'&'.join(result)).decode ("utf8")
	
	def unserialize_from_string(self, string):
		items = {}
		try:
			base64_hash, data = string.split(b'?', 1)
		except:
			base64_hash, data, items = b"", b"", {}
		else:
			mac = hmac(self.secret_key, None, self.hash_method)
			for item in data.split(b'&'):
				mac.update(b'|' + item)
				if not b'=' in item:
					items = None
					break
				key, value = item.split(b'=', 1)
				# try to make the key a string
				try:
					key = unquote_plus (key.decode ("utf8"))
				except UnicodeError:
					pass
				items[key] = value
			
			try:
				client_hash = base64.b64decode(base64_hash)
			except Exception:
				items = client_hash = None
			
			if items is not None and client_hash == mac.digest():
				try:
					for key, value in items.items():
						items[key] = self.unquote(value)
				except UnquoteError:
					items = {}
									
			else:
				items = {}
						
		self.data = items
		

#------------------------------------------------------
# List Type
#------------------------------------------------------

class SecuredListValue (SecuredValue):
	
	def set_default_data (self):
		self.data = []
	
	def serialize(self):
		if self.secret_key is None:
			raise RuntimeError('no secret key defined')
								
		result = []
		mac = hmac(self.secret_key, None, self.hash_method)		
		for value in sorted (self.data):
			result.append (self.quote (value))
			mac.update(b'|' + result[-1])
		return (base64.b64encode(mac.digest()).strip() + b"?" + b'&'.join(result)).decode ("utf8")
	
	def unserialize_from_string(self, string):
		items = []
		try:
			base64_hash, data = string.split(b'?', 1)
		except:
			base64_hash, data, items = b"", b"", []			
		else:
			try:
				client_hash = base64.b64decode(base64_hash)
			except Exception:
				items = client_hash = None
				
			mac = hmac(self.secret_key, None, self.hash_method)
			for item in data.split(b'&'):
				mac.update(b'|' + item)
				items.append (self.unquote (item))
			
			if client_hash != mac.digest():
				items = []
				
		self.data = items
		
				



import time
from aquests.lib.athreads import socket_map
from aquests.lib.athreads import trigger
import threading
from skitai.server.rpc import cluster_dist_call, rcache
from aquests.lib.attrdict import AttrDict
from aquests.dbapi import request
import asyncore
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB
from aquests.lib.cbutil import tuple_cb
		
class OperationTimeout (Exception):
	pass

class RequestFailed (Exception):
	pass

class FailedRequest:
	def __init__ (self, expt_class, expt_str):
		self.description = None
		self.data = None
		self.expt_class = expt_class
		self.expt_str = expt_str
		
		self.code, self.msg = 501, "timeout"

	def raise_for_status (self):
		if self.expt_class:
			raise self.expt_class (self.expt_str)
	reraise = raise_for_status

	
class Dispatcher:
	def __init__ (self, cv, id, ident = None, filterfunc = None, callback = None):
		self.__cv = cv
		self.id = id
		self.ident = ident
		self.filterfunc = filterfunc
		self.callback = callback
		self.creation_time = time.time ()
		self.status = 0
		self.result = None
			
	def get_id (self):
		return self.id
	
	def get_status (self):
		# 0: Not Connected
		# 1: Operation Timeout
		# 2: Exception Occured
		# 3: Normal
		self.__cv.acquire ()		
		status = self.status
		self.__cv.release ()
		return status
		
	def set_status (self, code):
		self.__cv.acquire ()
		self.status = code
		self.__cv.notify ()
		self.__cv.release ()
		return code
		
	def get_result (self):
		if self.result is None:
			if self.get_status () == -1:
				self.result = cluster_dist_call.Result (self.id, -1, FailedRequest (RequestFailed, "Request Failed"), self.ident)
			else:
				self.result = cluster_dist_call.Result (self.id, 1, FailedRequest (OperationTimeout, "Operation Timeout"), self.ident)
		return self.result
	
	def do_filter (self):
		if self.filterfunc:
			self.filterfunc (self.result)
						
	def handle_result (self, request):
		if self.get_status () == 1:
			# timeout, ignore
			return
				
		if request.expt_class:
			status = 2			
		else:
			status = 3
		
		self.result = cluster_dist_call.Result (self.id, status, request, self.ident)
		self.set_status (status)

		if self.callback:
			tuple_cb (self.result, self.callback)
			
				        	     
#-----------------------------------------------------------
# Cluster Base Call
#-----------------------------------------------------------

class ClusterDistCall (cluster_dist_call.ClusterDistCall):
	def __init__ (self, 
		cluster,
		server, 
		dbname, 
		auth,
		dbtype,
		meta,
		use_cache,
		mapreduce,
		filter,
		callback,
		timeout,
		logger
		):
		
		self.server = server
		self.dbname = dbname
		
		self.auth = auth		
		self.dbtype = dbtype
		if self.dbtype == DB_PGSQL:
			if self.dbname in (DB_SQLITE3, DB_REDIS):
				self.dbtype = self.dbname
				self.dbname = ""
			elif self.auth in (DB_MONGODB,):
				self.dbtype = self.auth
				self.auth = None
		self._meta = meta				
		self._use_cache = use_cache		
		self._filter = filter
		self._callback = callback
		self._timeout = timeout
		self._mapreduce = mapreduce		
		self._cluster = cluster
		self._cached_request_args = None
		self._cached_result = None
		
		self._logger = logger		
		self._requests = {}
		self._results = []
		self._canceled = 0
		self._init_time = time.time ()
		self._cv = None
		self._retry = 0	
		self._numnodes = 0
		self._sent_result = None
			
		if self._cluster:
			nodes = self._cluster.get_nodes ()
			self._numnodes = len (nodes)
			if self._mapreduce:
				self._nodes = nodes
			else: # anyone of nodes
				self._nodes = [None]
		
	def _get_ident (self):
		cluster_name = self._cluster.get_name ()
		if cluster_name == "dbpool":
			_id = "%s/%s/%s" % (self.server, self.dbname, self.auth)
		else:
			_id = cluster_name
		_id += "/%s%s" % (
			",".join (map (lambda x: str (x), self._cached_request_args)),
			self._mapreduce and "/M" or ""
			)	
		return _id
				
	def _get_connection (self, id = None):
		if id is None: id = self._nodes.pop ()
		else: self._nodes = []
			
		if self.server:
			asyncon = self._cluster.get (self.server, self.dbname, self.auth, self.dbtype)		
		else:	
			asyncon = self._cluster.get (id)		
		self._setup (asyncon)
		return asyncon
	
	def _request (self, method, params):
		# For Django QuerySet and SQLGen
		if hasattr (params [0], "query"):
			params = (str (params [0].query),) + params [1:]
			
		self._cached_request_args = (method, params) # backup for retry
		if self._use_cache and rcache.the_rcache:
			self._cached_result = rcache.the_rcache.get (self._get_ident (), self._use_cache)
			if self._cached_result is not None:
				return
		
		while self._avails ():
			asyncon = self._get_connection (None)			
			rs = Dispatcher (self._cv, asyncon.address, ident = not self._mapreduce and self._get_ident () or None, filterfunc = self._filter, callback = self._callback)
			self._requests [rs] = asyncon	
			
			req = request.Request (
				self.dbtype,
				self.server, 
				self.dbname,
				self.auth,
				method, params, 				
				rs.handle_result,
				self._meta
			)			
			asyncon.execute (req)			
		trigger.wakeup ()
		return self
	
	
class ClusterDistCallCreator:
	def __init__ (self, cluster, logger):
		self.cluster = cluster
		self.logger = logger
	
	def __getattr__ (self, name):	
		return getattr (self.cluster, name)
		
	def Server (self, server = None, dbname = None, auth = None, dbtype = None, meta = None, use_cache = True, mapreduce = False, filter = None, callback = None, timeout = 10):
		if cluster_dist_call.is_main_thread () and not callback:
			raise RuntimeError ('Should have callback in Main thread')
		return cluster_dist_call.Proxy (ClusterDistCall, self.cluster, server, dbname, auth, dbtype, meta, use_cache, mapreduce, filter, callback, timeout, self.logger)
		
		


from . import wsgi_handler
import skitai
from aquests.lib import producers
from h2.connection import H2Connection, GoAwayFrame, DataFrame, H2Configuration
from h2.exceptions import ProtocolError, NoSuchStreamError, StreamClosedError
from h2.events import DataReceived, RequestReceived, StreamEnded, PriorityUpdated, ConnectionTerminated, StreamReset, WindowUpdated, RemoteSettingsChanged
from h2.settings import SettingCodes
from h2.errors import ErrorCodes
from aquests.lib import producers
from .http2.request import request as http2_request
from .http2.vchannel import fake_channel, data_channel
from aquests.protocols.http2.producers import h2header_producer, h2frame_producer
from aquests.protocols.http2.fifo import http2_producer_fifo, http2_producer_ts_fifo
import threading
from io import BytesIO

class http2_request_handler:
	collector = None
	producer = None
	http11_terminator = 24
	def __init__ (self, handler, request):
		self.handler = handler
		self.wasc = handler.wasc
		self.request = request
		self.channel = request.channel
		
		# replace fifo with supporting priority, ready, removing
		if self.wasc.numthreads:
			self.channel.producer_fifo = http2_producer_ts_fifo ()
		else:	
			self.channel.producer_fifo = http2_producer_fifo ()
		
		self.conn = H2Connection(H2Configuration(client_side=False))
		self.frame_buf = self.conn.incoming_buffer
		self.frame_buf.max_frame_size = self.conn.max_inbound_frame_size
		
		self.data_length = 0
		self.current_frame = None
		self.rfile = BytesIO ()
		self.buf = b""

		self.requests = {}
		self.priorities = {}
		self.promises = {}
		
		self._send_stream_id = 0
		self._closed = False
		self._got_preamble = False		
		
		self._plock = threading.Lock () # for self.conn
		self._clock = threading.Lock () # for self.x
				
	def close (self, errcode = 0, msg = None):
		if self._closed: return
		self._closed = True		
		if self.channel:
			self.go_away (errcode) # go_away
		
		with self._clock:	
			stream_ids = list (self.requests.keys ())		
		for stream_id in stream_ids:
			self.remove_request (stream_id)
	
	def enter_shutdown_process (self):
		self.close (ErrorCodes.NO_ERROR)
			
	def closed (self):
		return self._closed
			
	def send_data (self):
		with self._plock:
			data_to_send = self.conn.data_to_send ()
		
		if data_to_send:
			#print ("SEND", repr (data_to_send), len (data_to_send))
			self.channel.push (data_to_send)
	
	def handle_preamble (self):
		if self.request.version.startswith ("2."):
			self.channel.set_terminator (6) # SM\r\n\r\n		
					
	def initiate_connection (self):
		self.handle_preamble ()
		h2settings = self.request.get_header ("HTTP2-Settings")
		if h2settings:
			self.conn.initiate_upgrade_connection (h2settings)
		else:	
			self.conn.initiate_connection()		
		self.send_data ()
		
		if self.request.version == "1.1":
			self.handle_request (1, self.upgrade_header ())			
			
	def upgrade_header (self):
		headers = [
			(":method", self.request.command.upper ()), 
			(":path", self.request.uri),
		]
		for line in self.request.get_header ():
			try:
				k, v = line.split (": ", 1) 
			except ValueError:
				k, v = line.split (":", 1) 	
			k = k.lower ()
			if k in ("http2-settings", "connection", "upgrade"):
				continue
			headers.append ((k, v))		
		return headers
					
	def collect_incoming_data (self, data):
		#print ("RECV", repr (data), len (data), '::', self.channel.get_terminator ())
		if not data:
			# closed connection
			self.close (True)
			return
			
		if self.data_length:
			self.rfile.write (data)
		else:
			self.buf += data
	
	def set_frame_data (self, data):
		if not self.current_frame:
			return []
		self.current_frame.parse_body (memoryview (data))				
		self.current_frame = self.frame_buf._update_header_buffer (self.current_frame)
		with self._plock:
			events = self.conn._receive_frame (self.current_frame)
		return events

	def set_terminator (self, terminator):
		self.channel.set_terminator (terminator)
					
	def found_terminator (self):
		buf, self.buf = self.buf, b""
		
		#print ("FOUND", repr (buf), '::', self.data_length, self.channel.get_terminator ())
		#print ("FOUND", self.request.version, self.request.command, self.data_length)
		events = None	
		if self.request.version == "1.1" and self.data_length:
			self.request.version = "2.0" # upgrade
			data = self.rfile.getvalue ()
			with self._clock:
				r = self.requests [1]				
			r.channel.set_data (data, len (data))
			r.set_stream_ended ()
			self.data_length = 0
			self.rfile.seek (0)
			self.rfile.truncate ()			
			self.channel.set_terminator (24) # for premble
		
		elif not self._got_preamble:			
			if not buf.endswith (b"SM\r\n\r\n"):
				raise ProtocolError ("Invalid preamble")
			self._got_preamble = True
			self.channel.set_terminator (9)
		
		elif self.data_length:
			events = self.set_frame_data (self.rfile.getvalue ())
			self.current_frame = None
			self.data_length = 0
			self.rfile.seek (0)
			self.rfile.truncate ()			
			self.channel.set_terminator (9) # for frame header
									
		elif buf:
			self.current_frame, self.data_length = self.frame_buf._parse_frame_header (buf)
			self.frame_buf.max_frame_size = self.data_length
			
			if self.data_length == 0:
				events = self.set_frame_data (b'')
			self.channel.set_terminator (self.data_length == 0 and 9 or self.data_length)	# next frame header
			
		else:
			raise ProtocolError ("Frame decode error")
		
		if events:
			self.handle_events (events)
	
	def get_new_stream_id (self):
		with self._clock:
			self._send_stream_id += 2
			stream_id = self._send_stream_id
		return stream_id
		
	def push_promise (self, stream_id, request_headers, addtional_request_headers):
		promise_stream_id = self.get_new_stream_id ()
		with self._clock:
			self.promises [promise_stream_id] = request_headers	+ addtional_request_headers	
		
		with self._plock:
			self.conn.push_stream (stream_id, promise_stream_id, request_headers	+ addtional_request_headers)
		self.send_data ()
					
	def handle_response (self, stream_id, headers, producer, do_optimize, force_close = False):
		r = self.get_request (stream_id)
		with self._clock:			
			try:
				depends_on, weight = self.priorities [stream_id]
			except KeyError:
				depends_on, weight = 0, 1	
			else:
				del self.priorities [stream_id]

		header_producer = h2header_producer (stream_id, headers, producer, self.conn, self._plock)
		
		if not producer:
			header_producer = r.response.log_or_not (r.uri, header_producer, r.response.log)			
			self.channel.push_with_producer (header_producer)
			
		else:
			self.channel.push_with_producer (header_producer)
			outgoing_producer = r.response.log_or_not (r.uri, producer, r.response.log)
			
			if do_optimize:
				outgoing_producer = producers.globbing_producer (outgoing_producer)
				
			outgoing_producer = h2frame_producer (
				stream_id, depends_on, weight, outgoing_producer, self.conn, self._plock
			)
			# is it proper?
			#outgoing_producer = producers.ready_globbing_producer (outgoing_producer)
			self.channel.push_with_producer (outgoing_producer)
		
		if r.is_stream_ended ():
			# needn't recv data any more
			self.remove_request (stream_id)
		
		if force_close:
			return self.go_away (ErrorCodes.CANCEL)
				
		current_promises = []
		with self._clock:
			while self.promises:
				current_promises.append (self.promises.popitem ())				
		for promise_stream_id, promise_headers in current_promises:
			self.handle_request (promise_stream_id, promise_headers, is_promise = True)

	def remove_request (self, stream_id):
		r = self.get_request (stream_id)
		if not r: return
		with self._clock:
			try: del self.requests [stream_id]
			except KeyError: pass		
		r.http2 = None
		
	def get_request (self, stream_id):
		r = None
		with self._clock:
			try: r =	self.requests [stream_id]
			except KeyError: pass
		return r
			
	def handle_events (self, events):
		for event in events:
			#print (event)
			if isinstance(event, RequestReceived):				
				self.handle_request (event.stream_id, event.headers)				
				
			elif isinstance(event, StreamReset):
				if event.remote_reset:
					r = self.get_request (event.stream_id)
					if r and r.collector:
						try: r.collector.stream_has_been_reset ()
						except AttributeError: pass
						r.set_stream_ended ()
						
					deleted = False
					if event.stream_id % 2 == 0: # promise stream
						with self._clock:
							try: del self.promises [event.stream_id]
							except KeyError: pass
							else: deleted = True

					if not deleted:
						self.channel.producer_fifo.remove (event.stream_id)
						
			elif isinstance(event, ConnectionTerminated):
				self.close (True)
					
			elif isinstance(event, PriorityUpdated):
				if event.exclusive:
					# rebuild depend_ons
					for stream_id in list (self.priorities.keys ()):
						depends_on, weight = self.priorities [stream_id]
						if depends_on == event.depends_on:
							self.priorities [stream_id] = [event.stream_id, weight]
					
				with self._clock:
					self.priorities [event.stream_id] = [event.depends_on, event.weight]
				
			elif isinstance(event, DataReceived):				
				r = self.get_request (event.stream_id)				
				if not r:
					self.go_away (ErrorCodes.PROTOCOL_ERROR)
				else:
					try:
						r.channel.set_data (event.data, event.flow_controlled_length)
					except ValueError:						
						# from vchannel.handle_read () -> collector.collect_inconing_data ()
						self.go_away (ErrorCodes.CANCEL)
					
			elif isinstance(event, StreamEnded):
				r = self.get_request (event.stream_id)
				if r:
					if r.collector:
						r.channel.handle_read ()
						r.channel.found_terminator ()					
					r.set_stream_ended ()
					if r.response.is_done ():
						# DO NOT REMOVE before responsing
						# this is for async streaming request like proxy request
						self.remove_request (event.stream_id)
					
		self.send_data ()
	
	def go_away (self, errcode = 0, msg = None):
		with self._plock:
			self.conn.close_connection (errcode, msg)
		self.send_data ()
		self.channel.close_when_done ()
	
	def reset_stream (self, stream_id, errcode = ErrorCodes.CANCEL):
		closed = False
		with self._plock:
			try:
				self.conn.reset_stream (stream_id, error_code = errcode)
			except StreamClosedError:	
				closed = True
		if not closed:		
			self.send_data ()
		self.remove_request (stream_id)
		#self.request.logger ("stream reset (stream_id:%d, error:%d)" % (stream_id, errcode), "info")
		
	def handle_request (self, stream_id, headers, is_promise = False):
		#print ("++REQUEST: %d" % stream_id, headers)
		command = "GET"
		uri = "/"
		scheme = "http"
		authority = ""
		cl = 0
		h = []
		cookies = []
		
		if type (headers [0][0]) is bytes:
			headers = [(k.decode ("utf8"), v.decode ("utf8")) for k, v in headers]
			
		for k, v in headers:
			#print ('HEADER:', k, v)								
			if k[0] == ":":
				if k == ":method": command = v.upper ()
				elif k == ":path": uri = v
				elif k == ":scheme": scheme = v.lower ()
				elif k == ":authority":
					authority = v
					if authority:
						h.append ("host: %s" % authority.lower ())				
				continue								
			if k == "content-length":
				cl = int (v)				
			elif k == "cookie":
				cookies.append (v)
				continue					
			h.append ("%s: %s" % (k, v))
								
		if cookies:
			h.append ("Cookie: %s" % "; ".join (cookies))
		
		should_have_collector = False	
		if command == "CONNECT":
			first_line = "%s %s HTTP/2.0" % (command, authority)
			vchannel = self.channel			
		else:	
			first_line = "%s %s HTTP/2.0" % (command, uri)
			if command in ("POST", "PUT"):
				should_have_collector = True
				vchannel = data_channel (stream_id, self.channel, cl)
			else:
				if stream_id == 1:
					self.request.version = "2.0"
				vchannel = fake_channel (stream_id, self.channel)

		r = http2_request (
			self,  scheme, stream_id, is_promise, 
			vchannel, first_line, command.lower (), uri, "2.0", h
		)		
		vchannel.current_request = r
		
		with self._clock:
			self.channel.request_counter.inc()
			self.channel.server.total_requests.inc()
		
		h = self.handler.default_handler
		if not h.match (r):
			try: r.response.error (404)			
			except: pass
			return	
			
		with self._clock:
			self.requests [stream_id] = r

		try:					
			h.handle_request (r)
			
		except:
			self.channel.server.trace()
			try: r.response.error (500)
			except: pass
				
		else:
			if should_have_collector and cl > 0:
				if r.collector is None:
					# POST but too large body or 3xx, 4xx
					if stream_id == 1 and self.request.version == "1.1":						
						self.channel.close_when_done ()
						self.remove_request (1)
					else:						
						self.go_away (ErrorCodes.PROTOCOL_ERROR)
				else:							
					if stream_id == 1 and self.request.version == "1.1":
						self.data_length = cl
						self.set_terminator (cl)

class h2_request_handler (http2_request_handler):
	http11_terminator = None
	
	def handle_preamble (self):
		if self.request.version.startswith ("2."):
			self.conn.receive_data ("PRI * HTTP/2.0\r\n\r\n")
			self.channel.set_terminator (None)
			
	def collect_incoming_data (self, data):		
		with self._plock:
			events = self.conn.receive_data (data)
		self.handle_events (events)


class Handler (wsgi_handler.Handler):
	keep_alive = 120
	
	def __init__(self, wasc, default_handler = None):
		wsgi_handler.Handler.__init__(self, wasc, None)
		self.default_handler = default_handler
		
	def match (self, request):
		return True
		
	def handle_request (self, request):
		is_http2 = False
		if request.command == "pri" and request.uri == "*" and request.version == "2.0":
			is_http2 = True
		else:	
			upgrade = request.get_header ("upgrade")		
			is_http2 = upgrade and upgrade.lower () == "h2c" and request.version == "1.1"

		if not is_http2:
			return self.default_handler.handle_request (request)
	
		http2 = http2_request_handler (self, request)		
		request.channel.die_with (http2, "http2 stream")
		request.channel.set_socket_timeout (self.keep_alive)
		
		if request.version == "1.1":
			request.response (
				"101 Switching Protocol",
				headers = [("Connection",  "upgrade"), ("Upgrade", "h2c"), ("Server", skitai.NAME.encode ("utf8"))]
			)
			request.response.done (upgrade_to = (http2, http2.http11_terminator))
		
		else:
			request.channel.current_request = http2
			
		http2.initiate_connection ()
	


'''# --------------------------------------------------------------------------
Events

def decorate (app):
    
    @app.on ("user:authenticated")
    def user_authenticated (was, form): pass
    
    @app.on ("user:added")
    def user_added (was, user, form): pass    
    
    @app.on ("user:updated")
    def user_updated (was, user, form): pass
    
     @app.on ("user:password:updated")
    def password_updated (was, user): pass
    
    @app.on ("user:password:reset-requested")
    def password_reset_requested (was, username, form): pass
    
# --------------------------------------------------------------------------'''

import re
from django.contrib.auth.password_validation import validate_password
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q

# Validators --------------------------------------------

RX_EMAIL = re.compile ('(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$')
def is_vaild_email (string):
    return RX_EMAIL.match (string)

def is_valid_length (string):
    return (4 < len (string) < 20)        

def is_valid_password (password):
    if not is_valid_length (password):
        return "Invalid password, required 4 - 20 characters"
    try:
        validate_password (password)
    except django.core.exceptions.ValidationError as why:
        return why.messages

def push_error_messages (was, msgs):
    if not msgs:
        return
    if isinstance (msgs, str):
        msgs = [msgs]
    for msg in msgs:
        was.mbox.push (msg, "error", icon = "new_releases")            

def update_password (was, username, password, template):
    errors = is_valid_password (password)
    if errors:
        push_error_messages (was, errors)
        return was.render (template)    
    user = User.objects.get (username = username)
    user.set_password (password)
    user.save ()    
    was.django.update_session_auth_hash (user)    
    was.app.emit ("user:password:updated", user)


# App Decorator ---------------------------------------------------------
        
def decorate (app):    
    # Handlers ---------------------------------------------------------        
    @app.login_handler
    def login_handler (was):
        if was.django.user.is_authenticated ():
            was.request.user = was.django.user
            was.app.emit ("user:authenticated", was.request.user)
            was.django.session.set_expiry (was.app.session_timeout) #extend timeout
            return
        
        if was.request.args.get ("_csrf_token"):
            if not was.csrf_verify ():
                return was.response ("400 Bad Request")
        
        next_url = not was.request.uri.endswith ("signout") and was.request.uri or ""
        return was.redirect (was.ab ("signin", next_url))
    
    @app.staff_member_check_handler
    def staff_member_check_handler (was):
        if not was.django.user.is_staff:
            return was.response ("403 Staff Permission Required")

    @app.permission_check_handler
    def permission_check_handler (was, perms):
        if isinstance (perms, str):
            perms = (prems,)
        if not was.django.user.has_perms (perms):            
            return was.response ("403 Permission Denied")    
    
    # Views ---------------------------------------------------------        
    @app.route ("/regist/signout")
    def signout (was):
        was.django.logout ()
        was.mbox.push ("Signed out successfully", "success")    
        return was.redirect (was.ab ('index'))
        
    @app.route ("/regist/signin")
    def signin (was, next_url = None, **form):
        if was.django.user.is_authenticated ():    
            return was.redirect (next_url)
        
        if form.get ("username"):
            user = was.django.authenticate (
                form ["username"], 
                form ["password"]
            )
            if user is not None:
                was.django.login (user)
                return was.redirect (next_url or was.ab ('index'))
            else:
                was.mbox.push ("Invalid user name or password", "error", icon = "new_releases")
        return was.render ("regist/signin.html", next_url = next_url or was.ab ("index"))

    @app.route ("/regist/signup")
    def signup (was, next_url = None, **form):
        def show_form (msg = None):
            push_error_messages (was, msg)
            return was.render ("regist/signup.html", next_url = next_url or was.ab ("index"), form = form)
        
        if was.django.user.is_authenticated ():    
            return was.redirect (next_url)
        
        if "username" not in form:
            return show_form ()
        
        # username ---------------------------------------------
        username = form.get ("username", "")
        if not is_valid_length (username):
            return show_form ("Invalid user ID, required 4 - 20 characters")
        try: user = User.objects.get (username = username)
        except ObjectDoesNotExist: pass
        else: return show_form ("User ID already exists, try anohter ID. please")
        
        # password  ---------------------------------------------
        errors = is_valid_password (form.get ("password"))
        if errors:
            return show_form (errors)
        
        # email  -------------------------------------------------
        if "email" in form and not is_vaild_email (form ["email"]):
            return show_form ("Invalid email address")
        if User.objects.filter (email__iexact = form ["email"]).count():        
            return show_form ("Email already exists, If you want recover your account <a href='{}'>click here</a>".format (was.ab ('forgot_password')))         
        
        user = User (username = form ["username"], email = form ["email"])
        user.set_password (form ["password"])
        user.save ()
        was.app.emit ('user:added', user, form)
        
        was.mbox.push ("Sign up success, thank you", "info")
        return signin (was, next_url, **form)
    
    @app.route ("/regist/account")
    @app.login_required
    def account (was, **form):
        if "email" not in form:
            return was.render ("regist/account.html", form = form)
        
        # email  -------------------------------------------------
        if "email" in form and not is_vaild_email (form ["email"]):
            was.mbox.push ("Invalid email", "error", icon = "new_releases")
            return was.render ("regist/account.html", form = form)
        if User.objects.filter (~Q (username__iexact = was.django.user.username), email__iexact = form ["email"]).count():        
            was.mbox.push ("Email already exists with another username", "error", icon = "new_releases")
            return was.render ("regist/account.html", form = form)

        user = User.objects.get (username = was.django.user.username)
        user.email = form ["email"]
        user.save ()
        was.app.emit ('user:updated', user, form)
        
        was.mbox.push ("Account updated successfully", "info")
        return was.redirect (was.ab ("index"))
    
    @app.route ("/regist/forgot-password")
    def forgot_password (was, **form):    
        if form.get ("email"):
            try:
                user = User.objects.filter (email = form ["email"]).order_by ("-id")[0]
            except IndexError:    
                was.mbox.push ("Email does not exists, check your email please", "error", icon = "new_releases")
                return was.render ('regist/forgot-password.html')
            was.app.emit ("user:password:reset-requested", user.username, form)
            was.mbox.push ("Email has been sent. check your email, please", "info")
            return was.redirect (was.ab ('index'))
        return was.render ('regist/forgot-password.html')

    @app.route ("/regist/reset-password")
    def reset_password (was, t, **form):
        if not t:
            # need token
            return was.response ("400 Bad Request")
        
        username = was.detoken (t)
        if not username:
            was.mbox.push ("Your token invalid or had been expired, request agian please", "error", icon = "new_releases")
            return was.render ('regist/forgot-password.html', username = username)
        
        if not form.get ("password"):
            return was.render ('regist/reset-password.html', username = username)
        
        nextform = update_password (was, username, form ["password"], 'regist/reset-password.html')
        if nextform:
            return nextform
        
        was.rmtoken (t)
        was.mbox.push ("Your password has been reset and changed", "info")
        return signin (was, was.ab ('index'), **{"username": username, "password":  form ["password"]})
        
    @app.route ("/regist/change-password")
    @app.login_required
    def change_password (was, **form):    
        if not form.get ("password"):
            return was.render ('regist/change-password.html')
        
        username = was.django.user.username
        nextform = update_password (was, username, form ["password"], 'regist/change-password.html')
        if nextform:
            return nextform
    
        was.mbox.push ("Your password has been changed", "info")
        return was.redirect (was.ab ('index'))
    


import re, sys
from functools import wraps
from urllib.parse import unquote_plus, quote_plus, urljoin
import os
from aquests.lib import importer, strutil
from types import FunctionType as function
import inspect
from skitai import was as the_was
import time, threading
from .storage import Storage
	
RX_RULE = re.compile ("(/<(.+?)>)")

class Part:
	def __init__ (self, *args, **kargs):			
		self.module = None
		self.packagename = None
		self.wasc = None				
		self._started = False
		self._reloading = False
		self.logger = None
		self.mount_p = "/"
		self.path_suffix_len = 0
		self.route_map = {}
		self.route_priority = []
		self.storage = Storage ()
		
		self.lock = threading.RLock ()
		self.init_time = time.time ()		
		self._decos = {}
		self._permission_map = {}
		self._jinja2_filters = {}
		self._template_globals = {}
		self._function_specs = {}
		self._function_names = {}
		self._conditions = {}
		self._need_authenticate = {True: [], False: []}		
		self._cond_check_lock = threading.RLock ()
		
		self._binds_server = [None] * 6
		self._binds_request = [None] * 4		
		self.handlers = {}
				
	def set_mount_point (self, mount):	
		if not mount:
			self.mount_p = "/"
		elif mount [-1] != "/":
			self.mount_p = mount + "/"
		else:
			self.mount_p = mount
		self.path_suffix_len = len (self.mount_p) - 1
				
	def init (self, module, packagename = "app", mount = "/"):
		self.module = module	
		self.packagename = packagename
		self.set_mount_point (mount)
		
		if self.module:
			self.abspath = self.module.__file__
			if self.abspath [-3:] != ".py":
				self.abspath = self.abspath [:-1]
			self.update_file_info	()
		
	def __getitem__ (self, k):
		return self.route_map [k]
	
	def log (self, msg):
		self.logger.log (msg)
	
	def trace (self):
		self.logger.trace ()
		
	def reload_package (self):
		importer.reloader (self.module)
		self.update_file_info	()
	
	def is_reloadable (self):
		if self.module is None: return False
		stat = os.stat (self.abspath)
		return self.file_info != (stat.st_mtime, stat.st_size)		
			
	def update_file_info (self):
		stat = os.stat (self.abspath)
		self.file_info = (stat.st_mtime, stat.st_size)
	
	def save_function_spec_for_routing (self, func):
		# save original function spec for preventing distortion by decorating wrapper
		# all wrapper has *args and **karg but we want to keep original function spec for auto parametering call
		if func.__name__ not in self._function_specs:
			# save origin spec
			self._function_specs [func.__name__] = inspect.getargspec(func)
	
	def get_function_spec_for_routing (self, func):
			return self._function_specs.get (func.__name__)
					
	#----------------------------------------------
	# App Life Cycling
	#----------------------------------------------
	def before_mount (self, f):
		self._binds_server [0] = f		
		return f
	start_up = before_mount
	startup = before_mount
	 
	def mounted (self, f):
		self._binds_server [3] = f
		return f
	
	def before_reload (self, f):
		self._binds_server [5] = f
		return f	
	onreload = before_reload
	reload = before_reload
	
	def reloaded (self, f):
		self._binds_server [1] = f
		return f
	
	def before_umount (self, f):
		self._binds_server [4] = f
		return f
	umount = before_umount
	
	def umounted (self, f):
		self._binds_server [2] = f
		return f
	shutdown = umounted
	
	# Request chains ----------------------------------------------
				
	def before_request (self, f):
		self._binds_request [0] = f
		return f
	
	def finish_request (self, f):
		self._binds_request [1] = f
		return f
	
	def failed_request (self, f):
		self._binds_request [2] = f
		return f
	
	def teardown_request (self, f):
		self._binds_request [3] = f
		return f
		
	# Auth ------------------------------------------------------
	
	def auth_required (self, f):	
		self._need_authenticate [True].append (f.__name__)
		return f
	
	def auth_not_required (self, f):
		self._need_authenticate [False].append (f.__name__)
		return f
		
	def login_handler (self, f):
		self._decos ["login_handler"] = f
		return f
	
	def login_required (self, f):
		self.save_function_spec_for_routing (f)
		@wraps(f)
		def wrapper (was, *args, **kwargs):
			_funcs = self._decos.get ("login_handler")
			if _funcs:				
				response = _funcs (was)
				if response is not None:
					return response
			return f (was, *args, **kwargs)
		return wrapper
	
	def staff_member_check_handler (self, f):
		self._decos ["staff_member_check_handler"] = f
		return f
	
	def staff_member_required (self, f):
		self.save_function_spec_for_routing (f)
		@wraps(f)
		def wrapper (was, *args, **kwargs):
			_funcs = self._decos.get ("staff_member_check_handler")
			if _funcs:
				response = _funcs (was)					
				if response is not None:
					return response
			return f (was, *args, **kwargs)
		return wrapper
	
	def permission_check_handler (self, f):
		self._decos ["permission_check_handler"] = f
		return f
	
	def permission_required (self, p):
		def decorator(f):
			self.save_function_spec_for_routing (f)
			self._permission_map [f] = isinstance (p, str) and [p] or p
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				_funcs = self._decos.get ("permission_check_handler")
				if _funcs:
					response = _funcs (was, self._permission_map [f])					
					if response is not None:
						return response
				return f (was, *args, **kwargs)
			return wrapper
		return decorator
	
	def testpass_required (self, testfunc):
		def decorator(f):
			self.save_function_spec_for_routing (f)			
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				response = testfunc (was)
				if response is False:
					return was.response ("403 Permission Denied")
				elif response is not True and response is not None:
					return response
				return f (was, *args, **kwargs)
			return wrapper
		return decorator
	
	# Automation ------------------------------------------------------	
	def preworks (self, *funcs):
		def decorator(f):
			self.save_function_spec_for_routing (f)
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				for func in funcs:
					response = func (was)
					if response is not None:
						return response
				return f (was, *args, **kwargs)
			return wrapper
		return decorator
	
	def postworks (self, *funcs):
		def decorator(f):
			self.save_function_spec_for_routing (f)			
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				for func in funcs:
					func (was)					
				return f (was, *args, **kwargs)
			return wrapper
		return decorator
	
	# Conditional Automation ------------------------------------------------------	
	def _check_condition (self, was, key, func, interval, mtime_func):
		now = time.time ()
		with self._cond_check_lock:
			oldmtime, last_check = self._conditions [key]
		
		if not interval or now - last_check > interval:
			mtime = mtime_func (key)					
			if mtime > oldmtime:
				response = func (was, key)
				with self._cond_check_lock:
					self._conditions [key] = [mtime, now]
				if response is not None:
					return response
					
			elif interval:
				with self._cond_check_lock:
					self._conditions [key][1] = now						
		
	def if_updated (self, key, func, interval = 1):
		def decorator(f):
			self.save_function_spec_for_routing (f)
			self._conditions [key] = [0, 0]
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				response = self._check_condition (was, key, func, interval, was.getlu)
				if response is not None:
					return response
				return f (was, *args, **kwargs)
			return wrapper
		return decorator
		
	def if_file_modified (self, path, func, interval = 1):
		def decorator(f):
			self.save_function_spec_for_routing (f)
			self._conditions [path] = [0, 0]
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				def _getmtime (path): 
					try:
						mtime = os.path.getmtime (path)
					except FileNotFoundError:
						return 0
					return mtime
					
				response = self._check_condition (was, path, func, interval, _getmtime)
				if response is not None:
					return response
				return f (was, *args, **kwargs)
			return wrapper
		return decorator
	
	# Websocket ------------------------------------------------------
	def websocket_config (self, spec, timeout = 60, onopen = None, onclose = None, encoding = "text"):
		def decorator(f):
			self.save_function_spec_for_routing (f)
			@wraps(f)
			def wrapper (was, *args, **kwargs):
				if not was.wshasevent ():
					return f (was, *args, **kwargs)
				if was.wsinit ():
					return was.wsconfig (spec, timeout, encoding)
				elif onopen and was.wsopened ():
					return onopen (was)
				elif onclose and was.wsclosed ():
					return onclose (was)
				
			return wrapper
		return decorator
	
	# Templaing -------------------------------------------------------
	
	def template_global (self, name):	
		def decorator(f):
			self.save_function_spec_for_routing (f)
			@wraps(f)
			def wrapper (*args, **kwargs):				
				return f (the_was._get (), *args, **kwargs)
			self._template_globals [name] = wrapper
			return wrapper
		return decorator
	
	def template_filter (self, name):	
		def decorator(f):
			self._jinja2_filters [name] = f
			@wraps(f)
			def wrapper (*args, **kwargs):				
				return f (*args, **kwargs)			
			return wrapper
		return decorator
		
	# Error handling ------------------------------------------------------
		
	def get_error_page (self, error):
		handler = self.handlers.get (error ['code'], self.handlers.get (0))
		if not handler:
			return
		was = the_was._get ()	
		# reset was.app for rendering
		was.app = self
		content =	handler [0] (was, error)
		was.app = None
		return content
		
	def add_error_handler (self, errcode, f, **k):
		self.handlers [errcode] = (f, k)		
		
	def error_handler (self, errcode, **k):
		def decorator(f):
			self.add_error_handler (errcode, f, **k)
			@wraps(f)
			def wrapper (*args, **kwargs):
				return f (*args, **kwargs)
			return wrapper
		return decorator
	
	def default_error_handler (self, f):
		self.add_error_handler (0, f)
		return f
	
	defaulterrorhandler = default_error_handler
	errorhandler = error_handler
	
	#----------------------------------------------
	# URL Building
	#----------------------------------------------		
	def url_for (self, thing, *args, **kargs):
		if thing.startswith ("/"):
			return self.basepath [:-1] + self.mount_p [:-1] + thing
		
		for func, name, fuvars, favars, numvars, str_rule, options in self.route_map.values ():
			if thing != name: continue			
			assert len (args) <= len (fuvars), "Too many params, this has only %d params(s)" % len (fuvars)						
			params = {}
			for i in range (len (args)):
				assert fuvars [i] not in kargs, "Collision detected on keyword param '%s'" % fuvars [i]
				params [fuvars [i]] = args [i]				
			
			for k, v in kargs.items ():
				params [k] = v
			
			url = str_rule
			if favars: #fancy [(name, type),...]. /fancy/<int:cid>/<cname>
				for n, t in favars:
					if n not in params:
						raise AssertionError ("Argument '%s' missing" % n)
					value = quote_plus (str (params [n]))
					if t == "string":
						value = value.replace ("+", "_")
					elif t == "path":
						value = value.replace ("%2F", "/")
					url = url.replace ("<%s%s>" % (t != "string" and t + ":" or "", n), value)
					del params [n]
			
			if params:
				url = url + "?" + "&".join (["%s=%s" % (k, quote_plus (str(v))) for k, v in params.items ()])
				
			return self.url_for (url)
	
	def build_url (self, thing, *args, **kargs):
		url = self.url_for (thing, *args, **kargs)
		if url:
			return url			
	
	#----------------------------------------------
	# Routing
	#----------------------------------------------						
	def route (self, rule, **k):
		def decorator (f):
			self.add_route (rule, f, **k)
			@wraps(f)
			def wrapper (*args, **kwargs):
				return f (*args, **kwargs)
			return wrapper
		return decorator
			
	def get_route_map (self):
		return self.route_map
	
	def set_route_map (self, route_map):
		self.route_map = route_map
									
	def try_rule (self, path_info, rule, rulepack):
		f, n, l, a, c, s, options = rulepack
		
		arglist = rule.findall (path_info)
		if not arglist: 
			return None, None
		
		arglist = arglist [0]
		if type (arglist) is not tuple:
			arglist = (arglist,)
			
		kargs = {}
		for i in range(len(arglist)):
			an, at = a [i]
			if at == "int":
				kargs [an] = int (arglist [i])
			elif at == "float":
				kargs [an] = float (arglist [i])
			elif at == "path":
				kargs [an] = unquote_plus (arglist [i])
			else:		
				kargs [an] = unquote_plus (arglist [i]).replace ("_", " ")
		return f, kargs
	
	def add_route (self, rule, func, **options):
		if not rule or rule [0] != "/":
			raise AssertionError ("Url rule should be starts with '/'")
		
		if not self._started and not self._reloading and func.__name__ in self._function_names:
			raise NameError ("Function <{}> is already defined".format (func.__name__))
		self._function_names [func.__name__] = None		
			
		fspec = self._function_specs.get (func.__name__) or inspect.getargspec(func)
		options ["args"] = fspec.args [1:]
		options ["varargs"] = fspec.varargs
		options ["keywords"] = fspec.keywords
		
		if fspec.defaults:
			defaults = {}
			argnames = fspec.args[(len (fspec.args) - len (fspec.defaults)):]
			for i in range (len (fspec.defaults)):
				defaults [argnames [i]] = fspec.defaults[i]
			options ["defaults"] = defaults
		
		s = rule.find ("/<")
		if s == -1:	
			self.route_map [rule] = (func, func.__name__, func.__code__.co_varnames [1:func.__code__.co_argcount], None, func.__code__.co_argcount - 1, rule, options)						
		else:
			s_rule = rule
			rulenames = []
			urlargs = RX_RULE.findall (rule)
			options ["urlargs"] = len (urlargs)
			for r, n in urlargs:
				if n.startswith ("int:"):
					rulenames.append ((n[4:], n[:3]))
					rule = rule.replace (r, "/([0-9]+)")
				elif n.startswith ("float:"):
					rulenames.append ((n[6:], n [:5]))
					rule = rule.replace (r, "/([.0-9]+)")
				elif n.startswith ("path:"):
					rulenames.append ((n[5:], n [:4]))
					rule = rule.replace (r, "/(.+)")	
				else:
					rulenames.append ((n, "string"))
					rule = rule.replace (r, "/([^/]+)")
			rule = "^" + rule + "$"			
			re_rule = re.compile (rule)				
			self.route_map [re_rule] = (func, func.__name__, func.__code__.co_varnames [1:func.__code__.co_argcount], tuple (rulenames), func.__code__.co_argcount - 1, s_rule, options)
			self.route_priority.append ((s, re_rule))
			self.route_priority.sort (key = lambda x: x [0], reverse = True)			
		
	def get_routed (self, method_pack):
		if not method_pack: 
			return
		temp = method_pack
		while 1:
			routed = temp [1]
			if type (routed) is not list:
				return routed
			temp = routed
					
	def route_search (self, path_info):
		if not path_info:
			return self.url_for ("/"), self.route_map ["/"]
		if path_info in self.route_map:
			return self.route_map [path_info][0], self.route_map [path_info]
		trydir = path_info + "/"
		if trydir in self.route_map:
			return self.url_for (trydir), self.route_map [trydir]
		raise KeyError
					
	def get_package_method (self, path_info, command, content_type, authorization, use_reloader = False):		
		if not (path_info.startswith (self.mount_p) or (path_info + "/").startswith (self.mount_p)):
			return None, None, None, None, None, 0
		
		path_info = path_info [self.path_suffix_len:]
		app, method, kargs, matchtype = self, None, {}, 0				
		
		# 1st, try find in self
		try:
			method, current_rule = self.route_search (path_info)
			
		except KeyError:
			for priority, rule in self.route_priority:
				current_rule = self.route_map [rule]
				method, kargs = self.try_rule (path_info, rule, current_rule)
				if method: 
					match = (rule, current_rule)
					matchtype = 2
					options = current_rule [-1]
					break
		
		else:			
			match = path_info
			if type (method) is not function:
				# object move
				matchtype = -1
			else:	
				matchtype = 1
				options = current_rule [-1]				
				
		if method is None:
			return None, None, None, None, None, 0
		
		if matchtype == -1: # 301 move
			return app, method, None, None, None, -1
				
		return app, [self._binds_request [0], method] + self._binds_request [1:4], kargs, options, match, matchtype
	
	#----------------------------------------------
	# Django Model
	#----------------------------------------------
	
	def _model_changed (self, sender, **karg):
		model_name = str (sender)[8:-2]
		karg ['x_model_class'] = model_name
		if 'created' not in karg:
			karg ["x_operation"] = 'D'
		elif karg["created"]:
			karg ["x_operation"] = 'C'
		else:
			karg ["x_operation"] = 'U'
		karg ["x_ignore"]	= True
		the_was._get ().setlu (model_name, sender, **karg)
	
	def redirect_signal (self, framework = "django"):
		if framework == "django":	
			from django.db.models.signals import post_save, post_delete		
			post_save.connect (self._model_changed)
			post_delete.connect (self._model_changed)
	model_signal = redirect_signal
			
	#----------------------------------------------
	# Starting App
	#----------------------------------------------
	PHASES = {
		'before_mount': 0,
		'mounted': 3,
		'before_reload': 5,
		'reloaded': 1,
		'before_umount': 4,
		'umounted': 2,		
	}
	def life_cycle (self, phase, obj):
		index = self.PHASES.get (phase)
		func = self._binds_server [index]
		if not func:
			return	
		try:
			func (obj)
		except:
			if self.logger:
				self.logger.trace ()
			else:
				raise							 

	def cleanup (self):
		# initing app & packages		
		pass
			
	def _start (self, wasc, route, reload = False):
		self.wasc = wasc
		if not route:
			self.basepath = "/"
		elif not route.endswith ("/"):			
			self.basepath = route + "/"
		else:
			self.basepath = route			
		
	def start (self, wasc, route):
		self.bus.emit ("app:starting", wasc)
		self._start (wasc, route)
		self.bus.emit ("app:started", wasc)
		self._started = True
		
	def restart (self, wasc, route):		
		self._reloading = True
		self.bus.emit ("app:restarting", wasc)	
		self._start (wasc, route, True)
		self.bus.emit ("app:restarted", wasc)	
		self._reloading = False



from skitai import lifetime
from aquests.lib import pathtool, logger
import os, signal, sys, tempfile
from aquests.lib.pmaster.processutil import set_process_name
from aquests.lib.pmaster import flock
import skitai
import time, sys, os
from _ast import arg

EXIT_CODE = None

class Daemon:
	NAME = "base"
	def __init__ (self, config, logpath, varpath, consol):
		self.config = config
		self.logpath = logpath
		self.varpath = varpath
		self.consol = consol
		self.last_maintern = 0
		self.flock = None
		self.shutdown_in_progress = False	
		self.handlers = {}	
		set_process_name ("%s: %s" % (config.get ("pname") or skitai.get_proc_title (), self.NAME))
		self.setup ()
		
	def maintern_shutdown_request (self, now):
		# for wind32 only, scheduled call by lifetime
		global EXIT_CODE
		
		req = self.flock.lockread ("signal")
		if not req: return
		self.logger ("%s: got signal - %s" % (self.NAME, req))
		if req in ("terminate", "kill"):
			EXIT_CODE = 0
		elif req == "restart":
			EXIT_CODE = 3	
		elif req == "rotate":
			try: self.logger.rotate ()
			except: self.logger.trace ()
		else:
			self.logger ("%s: unknown signal - %s" % (self.NAME, req), "error")
		self.flock.unlock ("signal")
		
		if EXIT_CODE is not None:
			self.handlers [req](None, None)
		
	def make_logger (self, create_flock = True):		
		self.logger = logger.multi_logger ()
		if self.consol:
			self.logger.add_logger (logger.screen_logger ())			
		if self.logpath:
			self.logger.add_logger (logger.rotate_logger (self.logpath, self.NAME, "weekly"))
			self.logger ("logging to {}".format (self.logpath), "info")		
		if create_flock and os.name == "nt":			
			self.flock = flock.Lock (os.path.join (self.varpath, "%s" % self.NAME))
			self.flock.unlockall ()
			
	def bind_signal (self, term, kill, hup):
		self.handlers ["terminate"] = term
		self.handlers ["kill"] = kill
		self.handlers ["restart"] = hup
		
		if os.name == "nt":
			signal.signal(signal.SIGBREAK, term)
		else:	
			def hUSR1 (signum, frame):	
				self.logger.rotate ()			
			
			signal.signal(signal.SIGUSR1, hUSR1)
			signal.signal(signal.SIGTERM, term)
			#signal.signal(signal.SIGKILL, kill)
			signal.signal(signal.SIGHUP, hup)
			
	def setup (self):
		raise NotImplementedError
			
					
def get_default_varpath ():
	from hashlib import md5
	fullpath = os.path.join (os.getcwd (), sys.argv [0])
	dpath, script = os.path.split (fullpath)
	if script.endswith (".py"):
		script = script [:-3]
	_var = 'skitai-%s-%s' % (script, md5 (dpath.encode ('utf8')).hexdigest ()[:8])
	return os.name == "posix" and '/var/tmp/%s' % _var or os.path.join (tempfile.gettempdir(), _var)

def make_service (service_class, config, logpath, varpath, consol):
	if logpath:
		pathtool.mkdir (logpath)
	if not varpath:	
		varpath = get_default_varpath ()		
	pathtool.mkdir (varpath)	
	
	lck = flock.Lock (os.path.join (varpath, "%s" % service_class.NAME))
	pidlock = lck.get_pidlock ()
	if pidlock.isalive ():
		print("[error] already running")
		sys.exit (0)
	
	if consol not in ("1", "yes"): # service mode
		from aquests.lib import devnull		
		sys.stdout = devnull.devnull ()		
		sys.stderr = open (os.path.join (varpath, "stderr-%s.log" % service_class.NAME), "a")
		
	return service_class (config, logpath, varpath, consol)


